	// NpSession.cpp: implementation of the CNpSession class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#if defined(WIN32)
#include <winsock2.h>
#endif
#include "NpSessionBase.h"
#include "NpPacketSender.h"
#include "NpSession.h"

//#include "GTLog.h"

#ifndef BIN_DUMP
#if _MSC_VER >= 1300
#define BIN_DUMP	__noop
#else
#define BIN_DUMP	0
#endif
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CNpSession::CNpSession()
{
	m_nFiringEventCount = 0;
	m_bDeleteMark = false;

	m_bUseProxySetting = false;
	m_pIckarSocket = NULL;
	
	m_pEvents = NULL;

	m_ePacketProcess = ePP_READY;
	m_pNcCurrentPacket = NULL;
	
	m_pReceiveBuffer = NULL;
	m_nReceiveBufferSize = 0;
	m_nCurrentReceiveLength = 0;
}

CNpSession::~CNpSession()
{
	if( m_pIckarSocket )
		m_pIckarSocket->Delete();
	
	if( m_pReceiveBuffer )
	{
		ASSERT( m_nReceiveBufferSize != 0 );
		delete [] m_pReceiveBuffer;
	}

	if( m_pNcCurrentPacket )
		m_pNcCurrentPacket->Delete();
	
	DLOG("CNpSession::~CNpSession() %p\n", this);

}

bool CNpSession::InitIckarSocket()
{
	if( m_pIckarSocket )
	{
		m_pIckarSocket->Close();

		if( m_bUseProxySetting )
		{
		}
	}
	else
	{
		m_pIckarSocket = _CreateIckarSocket(this);

		if( m_bUseProxySetting )
		{
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////////
// INpSession interface methods
//

bool CNpSession::AttachIckarSocket(IIckarSocket * pIckarSocket)
{
	if( m_pIckarSocket )
	{
		m_pIckarSocket->Delete();
	}
	
	m_pIckarSocket = pIckarSocket;
	m_pIckarSocket->SetEvents( this );

	return true;
}

IIckarSocket * CNpSession::DetachIckarSocket()
{
	if( m_pIckarSocket == NULL )
		return NULL;

	IIckarSocket * pIckarSocket;
	pIckarSocket = m_pIckarSocket;
	m_pIckarSocket = NULL;

	pIckarSocket->SetEvents( NULL );
	return pIckarSocket;
}

void CNpSession::SetEvents(_INpSessionEvents * pEvents)
{
	m_pEvents = pEvents;
}

IIckarSocket * CNpSession::GetIckarSocket()
{
	if( m_pIckarSocket == NULL )
		InitIckarSocket();

	return m_pIckarSocket;
}

bool CNpSession::ConnectSession(LPCSTR szHostAddr, int nHostPort, bool bUseProxySetting, int nTimeout)
{
	bool bRes;
	DLOG("[ENTER] CNpSession::ConnectSession(szHostAddr=%s, nHostPort=%d) m_pIckarSocket=%p, this=%p", szHostAddr, nHostPort, m_pIckarSocket, this );
	
	if( m_pIckarSocket && m_pIckarSocket->IsConnected() )
	{
//		GT_FLOG(L"Already socket is connected");
		m_pIckarSocket->Close();
		return false;
	}

	m_bUseProxySetting = bUseProxySetting;
	InitIckarSocket();

	bRes = m_pIckarSocket->Connect( szHostAddr, nHostPort, nTimeout);
	DLOG("[LEAVE] CNpSession::ConnectSession(szHostAddr=%s, nHostPort=%d) m_pIckarSocket=%p, this=%p\n", szHostAddr, nHostPort, m_pIckarSocket, this );

	//GT_FLOG(L"ConnectSession return %s", bRes ? L"TRUE" : L"FALSE");

	return bRes;
}

void CNpSession::CloseSession()
{
	DLOG("[ENTER] CNpSession::CloseSession() m_pIckarSocket=%p, this=%p", m_pIckarSocket, this );
	
	m_nCurrentReceiveLength = 0;
	m_ePacketProcess = ePP_READY;

	if( m_pReceiveBuffer )
	{
		ASSERT( m_nReceiveBufferSize != 0 );
		delete [] m_pReceiveBuffer;

		m_nReceiveBufferSize = 0;
		m_pReceiveBuffer = NULL;
	}

	if( m_pIckarSocket )
		m_pIckarSocket->Close();

	DLOG("[LEAVE] CNpSession::CloseSession() m_pIckarSocket=%p, this=%p\n", m_pIckarSocket, this );
}

void CNpSession::SendRawData(LPSTR pBuffer, int nDataLen)
{
	if( m_pIckarSocket )
	{
		m_pIckarSocket->Send(pBuffer, nDataLen);
	}
}

int CNpSession::RecvRawData(LPSTR pBuffer, int nDataLen)
{
	if( m_pIckarSocket )
	{
		return m_pIckarSocket->Receive(pBuffer, nDataLen);
	}

	return 0;
}


bool CNpSession::SendNpPacket(const INpPacket * pPacket)
{
	return CNpPacketSender::SendNpPacket( m_pIckarSocket, pPacket );
}

bool CNpSession::SendNpPacket_M6V2(USHORT cmmd, ULONG trid, USHORT usBodyDataLength, LPCSTR szBodyData, LPC_NP_UTF8 szTo, LPC_NP_UTF8 szFrom, USHORT usSessionId)
{
	return CNpPacketSender::SendNpPacket_M6V2( m_pIckarSocket, cmmd, trid, usBodyDataLength, szBodyData, 0, szTo, szFrom, usSessionId );
}

bool CNpSession::SendNpPacket_M6V2(const M6V2_HEADER & header, USHORT usBodyDataLength, LPCSTR szBodyData, LPC_NP_UTF8 szTo, LPC_NP_UTF8 szFrom)
{
	return CNpPacketSender::SendNpPacket_M6V2( m_pIckarSocket, header, usBodyDataLength, szBodyData, szTo, szFrom );
}

bool CNpSession::SendNpPacket_MediaData(USHORT trid, USHORT usBodyDataLength, LPCSTR szBodyData)
{
	return CNpPacketSender::SendNpPacket_MediaData( m_pIckarSocket, trid, usBodyDataLength, szBodyData );
}

bool CNpSession::SendNpPacket_MediaPing(USHORT trid)
{
	return CNpPacketSender::SendNpPacket_MediaPing( m_pIckarSocket, trid );
}

bool CNpSession::SendNpPacket_SessionControlPacket(UCHAR data, bool checkbit)
{
	if( m_pIckarSocket == NULL )
		return false;

	return CNpPacketSender::SendNpPacket_SessionControlPacket( m_pIckarSocket, data, checkbit );
}


void CNpSession::Delete()
{
	DLOG("[ENTER] CNpSession::Delete() %p %d thread = %p", this, m_nFiringEventCount, GetCurrentThreadId());

	if( m_nFiringEventCount == 0 )
        delete this;
	else
		m_bDeleteMark = true;

	DLOG("[LEAVE] CNpSession::Delete() %p\r\n", this);
}

bool CNpSession::SendNpPacket_M6V2(USHORT cmmd, ULONG trid, USHORT usBodyDataLength, LPCSTR szBodyData, BYTE ucEncType, LPC_NP_UTF8 pu8_To, LPC_NP_UTF8 pu8_From, USHORT usSessionId)
{
	return CNpPacketSender::SendNpPacket_M6V2( m_pIckarSocket, cmmd, trid, usBodyDataLength, szBodyData, ucEncType, pu8_To, pu8_From, usSessionId );
}


////////////////////////////////////////////////////////////////////////////////////
// receive events : _IIckarSocketEvents
// 
void CALLBACK CNpSession::OnIckarSocket_Connect(IIckarSocket * pIckarSocket, int nSocketError)
{
	DLOG("[ENTER] CNpSession::OnIckarSocket_Connect(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );

	fire_OnConnect( nSocketError );
	if( Check_AsyncDelete() )
		return;

	DLOG("[LEAVE] CNpSession::OnIckarSocket_Connect(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
}

void CALLBACK CNpSession::OnIckarSocket_Close(IIckarSocket * pIckarSocket, int nSocketError)
{
	DLOG("[ENTER] CNpSession::OnIckarSocket_Close(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );

	fire_OnClose( nSocketError, false );
	if( Check_AsyncDelete() )
		return;

	DLOG("[LEAVE] CNpSession::OnIckarSocket_Close(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
}	

void CALLBACK CNpSession::OnIckarSocket_Accept(IIckarSocket * pIckarSocket, int nSocketError)
{
	DLOG("[ENTER] CNpSession::OnIckarSocket_Accept(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
	
	ASSERT( false );

	DLOG("[LEAVE] CNpSession::OnIckarSocket_Accept(pIckarSocket=%p, nError=%d), this=%p", pIckarSocket, nSocketError, this );
}

void CALLBACK CNpSession::OnIckarSocket_SendReady(IIckarSocket * pIckarSocket)
{	
	fire_OnSendReady();
	if( Check_AsyncDelete() )
		return;
}

void CALLBACK CNpSession::OnIckarSocket_ReceiveReady(IIckarSocket * pIckarSocket, int nBytes)
{
	//return ;
	// Ŷ Ľ ȿ ̱ ؼ 2  ۸ Ѵ.
	// 1)  ۸  Ͱ  ׳ stack  ۸   ͸ о̰ ĽѴ.
	// 2)  Ŷ  Ȯ  ʾ,, ۸ ؼ  Ϳ  Ľ ؾ  
	//    ɹ  ۰ Ҵؼ    װ ͸ ̾ ޾Ƽ ٽ ĽѴ.
	// κ  1 ǹǷ ޸  ϰ,, new/delete   带
	//    ̴.
	
	if( ePP_READY == m_ePacketProcess )
	{
		//  ۸ εͰ Ƿ stack  ӽ ۸  Ѵ.
		ASSERT( m_pReceiveBuffer == NULL );
		ASSERT( m_nReceiveBufferSize == 0 );
		ASSERT( m_nCurrentReceiveLength == 0 );

		char szRecvBuff[NPPACKET_MAX_HEADER_LENGTH+1];
		int n = m_pIckarSocket->Receive( szRecvBuff, NPPACKET_MAX_HEADER_LENGTH );
		//DLOG( "[call] m_pIckarSocket->Receive(), n=%d", n );
		if( n <= 0 )
		{
			if( n < 0 && m_pIckarSocket->GetLastWouldBlockError() )
				return;

			fire_OnClose(m_pIckarSocket->GetLastError(), false);
			Check_AsyncDelete();
			return;
		}

		szRecvBuff[n] = 0;
		
		const int nRemainDataSize = ParseSessionStream( szRecvBuff, n );
		if( nRemainDataSize < 0 )
		{
			// parsing  fail  ų. event  ޾ event receiver   Close() 
			return;
		}

		//  Ͱ ־   ̹  ͸ ؾ   new  Ҵ ۸ Ѵ.
		//
		switch( m_ePacketProcess )
		{
		case ePP_READY:
			// ּ 1 Ŷ Ľ̵Ǿ dispatch Ǿ  ʹ .
			ASSERT( nRemainDataSize == 0 );
			if( m_pNcCurrentPacket )
			{
				m_pNcCurrentPacket->Delete();
				m_pNcCurrentPacket = NULL;
			}
			return;

		case ePP_WAIT_HEADER_END:
			if( nRemainDataSize >= NPPACKET_MAX_HEADER_LENGTH )
			{
				// header  max ̰ Ǿµ Ľ̿  ߴ... invalid packet ,,  ó ʿ..
				CloseSession();
				
				fire_OnClose(0, true);			
				Check_AsyncDelete();
				return;
			}

			m_nReceiveBufferSize = NPPACKET_MAX_HEADER_LENGTH;
			break;

		case ePP_WAIT_BODY_END:
			m_nReceiveBufferSize = m_pNcCurrentPacket->BodyLength();
			if( m_nReceiveBufferSize < NPPACKET_MAX_HEADER_LENGTH )
				m_nReceiveBufferSize = NPPACKET_MAX_HEADER_LENGTH;
			break;

		default:
			ASSERT( false );
			return;
		}
		
		ASSERT( m_pReceiveBuffer == NULL );
		m_pReceiveBuffer = new char[m_nReceiveBufferSize + 1];
		m_pReceiveBuffer[0] = 0;
		m_nCurrentReceiveLength = 0;

		if( nRemainDataSize > 0 )
		{
			int nTotalProcessedLen = n - nRemainDataSize;
			memcpy( m_pReceiveBuffer, szRecvBuff + nTotalProcessedLen, nRemainDataSize );
			m_nCurrentReceiveLength = nRemainDataSize;
			m_pReceiveBuffer[ m_nCurrentReceiveLength ] = 0;
		}
	}
	else
	{
		// KEH_API
		// ̰  ϴ ذ Ȱ.
		// ASSERT  return !
		// ݵƮž DEBUG ڲ ⼭ ASSERT .
		if(m_pReceiveBuffer == NULL)
			return;

		// m_ePacketProcess == ( ePP_READY or ePP_WAIT_BODY_END )
		// m_pReceiveBuffer  ݵ ҴǾ ־ Ѵ.  ̰ 0 ,,
		ASSERT( m_pReceiveBuffer != NULL );

		int n = m_pIckarSocket->Receive( m_pReceiveBuffer + m_nCurrentReceiveLength,
			m_nReceiveBufferSize - m_nCurrentReceiveLength );
//		DLOG( "[call] m_pIckarSocket->Receive(), n=%d, m_nCurrentReceiveLength=%d", n, m_nCurrentReceiveLength );
//		DLOG( "[call] m_pIckarSocket->Receive(), n=%d", n );
		if( n <= 0 )
		{
#if defined(WIN32)
			int nError = m_pIckarSocket->GetLastError();
			if( n < 0 && WSAEWOULDBLOCK == nError )
				return;
#else
			int nError = 0;
#endif			
			fire_OnClose(nError, false);			
			Check_AsyncDelete();
			return;
		}
		
		m_nCurrentReceiveLength += n;
		m_pReceiveBuffer[ m_nCurrentReceiveLength ] = 0;
		
		const int nRemainDataSize = ParseSessionStream( m_pReceiveBuffer, m_nCurrentReceiveLength );
		if( nRemainDataSize < 0 )
		{
			// parsing  fail  ų. event  ޾ event receiver   Close() 
			return;
		}
		const int nProcessedDataSize = m_nCurrentReceiveLength - nRemainDataSize;

		//  Ͱ  ..
		int nRequiredBufferSize = NPPACKET_MAX_HEADER_LENGTH;
		switch( m_ePacketProcess )
		{
		case ePP_READY:
			ASSERT( nRemainDataSize == 0 );
			
			if( m_pNcCurrentPacket )
			{
				m_pNcCurrentPacket->Delete();
				m_pNcCurrentPacket = NULL;
			}

			delete [] m_pReceiveBuffer;
			m_pReceiveBuffer = NULL;
			m_nReceiveBufferSize = 0;
			m_nCurrentReceiveLength = 0;
			return;

		case ePP_WAIT_HEADER_END:
			if( nRemainDataSize >= NPPACKET_MAX_HEADER_LENGTH )
			{
				// header  max ̰ Ǿµ Ľ̿  ߴ... invalid packet ,,  ó ʿ..
				CloseSession();
				
				fire_OnClose(0, true);			
				Check_AsyncDelete();
				return;
			}
			break;
		case ePP_WAIT_BODY_END:
			// body data  ް ִ ̴.
			ASSERT( nRemainDataSize >= 0 && nRemainDataSize <= 65535 );

			if( m_pNcCurrentPacket->BodyLength() > nRequiredBufferSize )
				nRequiredBufferSize = m_pNcCurrentPacket->BodyLength();
			break;

		default:
			ASSERT( false );
			return;
		}

		ASSERT( nRequiredBufferSize > nRemainDataSize );
		
		if( m_nCurrentReceiveLength == nRemainDataSize )
		{
			// ParseSessionStream()  ó Ͱ  . -   ۾ ʿġ ʴ.
			return;
		}			
		ASSERT( m_nCurrentReceiveLength > nRemainDataSize );
		
		
		// Ϻ ó Ͱ ְ ִ ͵ ִ.
		//    Ѵ.
		if( m_nReceiveBufferSize == nRequiredBufferSize &&
			(m_nCurrentReceiveLength - nRemainDataSize) > nRemainDataSize ) 
		{
			// ʿ  ũⰡ   ũ ġ Ѱ
			//  ۿ     ũⰡ   ũ  ũ.
			memcpy( m_pReceiveBuffer, m_pReceiveBuffer+nProcessedDataSize, nRemainDataSize );
			m_nCurrentReceiveLength = nRemainDataSize;
		}
		else
		{
			char * pTmpBuffer = new char[nRequiredBufferSize + 1];

			memcpy( pTmpBuffer, m_pReceiveBuffer+nProcessedDataSize, nRemainDataSize );
			delete [] m_pReceiveBuffer;

			m_pReceiveBuffer = pTmpBuffer;
			m_nReceiveBufferSize = nRequiredBufferSize;
			m_nCurrentReceiveLength = nRemainDataSize;
			m_pReceiveBuffer[m_nCurrentReceiveLength] = 0;
		}
	}

}
/*
//static 
// return value : -1 (⺻ ش  ŭ  ޾..)
//				  -2 (⺻ ش  ŭ ޾Ұ,, remain   .
inline int sParseNpPacketHeader( CNpPacket * pNpPacket, LPCSTR pszRawData, int nLen )
{
	USHORT usHeaderLength = 0;
	
	if( pszRawData[0] == M6V2_MAGIC_CODE )
	{
		if( nLen < sizeof( M6V2_HEADER ) )
			return -1;
		
		usHeaderLength = ((M6V2_HEADER *)pszRawData)->usHeaderLength;

		if( nLen < usHeaderLength )
			return -2;
	}
	else if( IS_MEDIA_PACKET( pszRawData[0] ) )
	{
		if( nLen < MEDIA_PACKET_HEADER_SIZE )
			return -1;

		usHeaderLength = MEDIA_PACKET_HEADER_SIZE;
	}
	else if( IS_SESSION_CONTROL_PACKET( pszRawData[0] ) )
	{
		usHeaderLength = 1;
	}
	else 
	{
		return -3;	// Invalid Packet
	}

	pNpPacket->AssignRawHeader( (const BYTE*)pszRawData, usHeaderLength );

	return usHeaderLength;
}
*/
int CNpSession::ParseSessionStream( char * pszRawData, int nLen )
{
	if( ePP_WAIT_BODY_END != m_ePacketProcess )
	{	
		int nPreProcessedLen = fire_OnRecvRawData(pszRawData, nLen);
		if( Check_AsyncDelete() )
			return -1;

		if( nPreProcessedLen >= nLen )
		{
			ASSERT( nPreProcessedLen == nLen );
			return 0;
		}

		if( m_pIckarSocket == NULL || !m_pIckarSocket->IsConnected() )
		{
			ASSERT( m_ePacketProcess == ePP_READY );
			return -1;
		}

		pszRawData += nPreProcessedLen;
		nLen -= nPreProcessedLen;

		if( !m_pNcCurrentPacket )
			m_pNcCurrentPacket = new CNpPacket;
		else
			m_pNcCurrentPacket->Clear();
		
		m_ePacketProcess = ePP_WAIT_HEADER_END;

		ASSERT( nLen >= 1 );
		if( nLen < 1 )
			return nLen;
		
		int nHeaderLength = CNpPacket::sParseNpPacketHeader(m_pNcCurrentPacket, pszRawData, nLen);
		if( nHeaderLength == -1 )
			return nLen;		// ⺻ ش  ŭ ޾,,  .. ( ȸ~)
		
		if( nHeaderLength == -2 )
			return nLen;

		if( nHeaderLength == -3 )
		{
			fire_OnClose(0, true);			
			Check_AsyncDelete();
			return -1;
		}
		
		if( m_pNcCurrentPacket->BodyLength() == 0 )
		{
			// header only packet
			DLOG("[EVENT CALL] fire_OnRecvPacket(), this=%p", this);
			NPPACKET_DUMP( m_pNcCurrentPacket );
			fire_OnRecvPacket();
			if( Check_AsyncDelete() )
				return -1;
			DLOG("[EVENT RETURN] fire_OnRecvPacket(), this=%p\r\n", this);

			if( m_pIckarSocket == NULL || !m_pIckarSocket->IsConnected() )
			{
				ASSERT( m_ePacketProcess == ePP_READY );
				return -1;
			}			
			
			m_ePacketProcess = ePP_READY;
		}
		else
		{
			m_ePacketProcess = ePP_WAIT_BODY_END;
		}
		
		int nRemainDataSize = nLen - nHeaderLength;
		
		if( nRemainDataSize > 0 )
			return ParseSessionStream( pszRawData + (nLen - nRemainDataSize), nRemainDataSize );
	}
	else // ePP_WAIT_BODY_END == m_ePacketProcess
	{
		ASSERT( m_pNcCurrentPacket != NULL );

		if( nLen < m_pNcCurrentPacket->BodyLength() )
		{
			return nLen;
		}
		else
		{
			m_pNcCurrentPacket->m_pBodyData = pszRawData;

			DLOG("[EVENT CALL] fire_OnRecvPacket(), this=%p", this);
			NPPACKET_DUMP( m_pNcCurrentPacket );
			fire_OnRecvPacket();
			if( Check_AsyncDelete() )
				return -1;
			DLOG("[EVENT RETURN] fire_OnRecvPacket(), this=%p\r\n", this);

			if( m_pIckarSocket == NULL || !m_pIckarSocket->IsConnected() )
			{
				ASSERT( m_ePacketProcess == ePP_READY );
				return -1;
			}

			const int nRemainDataSize = nLen - m_pNcCurrentPacket->BodyLength();

			m_pNcCurrentPacket->m_pBodyData = NULL;
			m_pNcCurrentPacket->Clear();
			m_ePacketProcess = ePP_READY;

			if( nRemainDataSize > 0 )
				return ParseSessionStream( pszRawData + (nLen - nRemainDataSize), nRemainDataSize );
		}
	}

	return 0;
}
